import json
import pickle
import io
import sys

import base64
from pprint import pprint

# This is a UUID we use to demarcate JSON data that needs specialized
# deserialization logic.  Because it's a UUID, we can expect it not to
# be present in user objects.
DMN_JSON = "47ce6720-cd47-44c5-a46f-86e346fc384f"

# TODO: all other possible types

def gm(name):
    mod = sys.modules.get(name, None)
    if mod is None:
        raise RuntimeError(f"Module {name} is not available")
    return mod

def binary_serialize(f):
    def inner(obj):
        tmp_io = io.BytesIO()
        res = f(obj, tmp_io)
        tmp_io.seek(0)
        return {
            'metadata': res,
            'b64': base64.b64encode(tmp_io.read()).decode('utf-8'),
        }
    return inner

def binary_deserialize(f):
    def inner(obj):
        tmp_io = io.BytesIO(base64.b64decode(obj['b64'].encode('utf-8')))
        return f(tmp_io, obj['metadata'])
    return inner

@binary_serialize
def serialize_numpy_ndarray(obj, io_out):
    gm('numpy').savez_compressed(io_out, array=obj)

@binary_deserialize
def deserialize_numpy_ndarray(io_in, _):
    return gm('numpy').load(io_in)["array"]

@binary_serialize
def serialize_torch(obj, io_out):
    gm('torch').save(obj, io_out)

@binary_deserialize
def deserialize_torch(io_in, _):
    return gm('torch').load(io_in)

@binary_serialize
def serialize_tensorflow_tensor(obj, io_out):
    tf = gm('tensorflow')
    if not hasattr(tf, "Session"):
        tf = tf.compat.v1
    sess = tf.get_default_session()
    if sess is None:
        sess = tf.Session()

    io_out.write(tf.serialize_tensor(obj).eval(session=sess))

    return {'out_type': obj.dtype.name}

@binary_deserialize
def deserialize_tensorflow_tensor(io_in, metadata):
    tf = gm('tensorflow')
    if not hasattr(tf, "Session"):
        tf = tf.compat.v1
    return tf.io.parse_tensor(io_in.read(), **metadata)

@binary_serialize
def serialize_tensorflow_eager_tensor(obj, io_out):
    tf = gm('tensorflow')
    if not hasattr(tf, "Session"):
        tf = tf.compat.v1
    io_out.write(tf.serialize_tensor(obj).numpy())
    return {'out_type': obj.dtype.name}

@binary_deserialize
def deserialize_tensorflow_eager_tensor(io_in, metadata):
    tf = gm('tensorflow')
    if not hasattr(tf, "Session"):
        tf = tf.compat.v1
    return tf.io.parse_tensor(io_in.read(), **metadata)

def serialize_bytes(obj):
    return base64.b64encode(obj).decode('utf-8')

def deserialize_bytes(s):
    return base64.b64decode(s)

def serialize_pandas_df(obj):
    # This will preserve wonky indexes
    return obj.to_json(orient='split')

def deserialize_pandas_df(s):
    return gm('pandas').read_json(s, orient='split')

FORMAT_MAP = {
    'numpy.ndarray': [serialize_numpy_ndarray, deserialize_numpy_ndarray],
    'torch.Tensor': [serialize_torch, deserialize_torch],
    'tensorflow.python.framework.ops.Tensor': [
        serialize_tensorflow_tensor, deserialize_tensorflow_tensor],
    'tensorflow.python.framework.ops.EagerTensor': [
        serialize_tensorflow_eager_tensor, deserialize_tensorflow_eager_tensor],
    'pandas.core.frame.DataFrame': [serialize_pandas_df, deserialize_pandas_df],
    'builtins.bytes': [serialize_bytes, deserialize_bytes]
}

def custom_attempt_deserialize(obj):
    if DMN_JSON in obj:
        qualname = obj[DMN_JSON][0]
        json = obj[DMN_JSON][1]

        if isinstance(qualname, str):
            func = FORMAT_MAP[qualname][1]
            return func(json)
        elif qualname is None:
            # Fallback case we're potentially loading pickled data
            return pickle.loads(base64.b64decode(json.encode('utf-8')))
        else:
            raise RuntimeError("Somehow qualname was neither a str nor None")
    return obj

def pickle_unsafe(obj):
    obj_data = gm('cloudpickle').dumps(obj)
    return {
        DMN_JSON: [None, base64.b64encode(obj_data).decode('utf-8')],
    }

def dumps(obj, serialization_fallback=None):
    def custom_attempt_serialize(obj):
        qualname = f"{obj.__class__.__module__}.{obj.__class__.__qualname__}"
        # This is only called if we failed the default deserializer
        if qualname in FORMAT_MAP:
            return {
                DMN_JSON: [qualname, FORMAT_MAP[qualname][0](obj)],
            }
        else:
            if serialization_fallback is not None:
                return serialization_fallback(obj)
            else:
                raise RuntimeError(f'Unable to serialize `{qualname}`.  Please specify a `serialization_fallback`, such as `dmn.json_utils.pickle_unsafe`.')

    return json.dumps(obj, default=custom_attempt_serialize)

def loads(s):
    # handle custom objects
    return json.loads(s, object_hook=custom_attempt_deserialize)



import cloudpickle
import sys
import re
import importlib_metadata
import inspect
import base64
import os

_required_dists = set([])
_old_is_registered_pickle_by_value = cloudpickle.cloudpickle._is_registered_pickle_by_value

PKG_TO_DIST = None
def get_dist_name(module_name):
    global PKG_TO_DIST
    if PKG_TO_DIST is None:
        PKG_TO_DIST = importlib_metadata.packages_distributions()
    lst = PKG_TO_DIST.get(module_name, [])
    if len(lst) < 1:
        return None
    return lst[0]

def _is_registered_pickle_by_value(module):
    builtin_regexes = [
        '^.*lib/python[.0-9]+/%s/.*$',
        '^.*lib/python[.0-9]+\\.zip/%s/.*$',
        '^.*lib/python[.0-9]+/lib-dynload/%s/.*$',
        '^.*lib/python[.0-9]+/%s\\.py$',
        '^.*lib/python[.0-9]+\\.zip/%s\\.py$',
        '^.*lib/python[.0-9]+/lib-dynload/%s\\.py$',
    ]
    site_packages_regexes = [
        '^.*lib/python3\\.[.0-9]+/site-packages/%s/.*$',
        '^.*lib/python3\\.[.0-9]+/dist-packages/%s/.*$',
    ]
    module_rootname = module.__name__.split('.')[0]
    module_path = getattr(module, '__file__', None)

    res = _old_is_registered_pickle_by_value(module)
    if module_path is None:
        return res
    elif any(re.match(r % module_rootname, module_path) for r in builtin_regexes):
        return res
    elif any(re.match(r % module_rootname, module_path) for r in site_packages_regexes):
        if not res:
            try:
                dist_name = get_dist_name(module_rootname)
                if dist_name is None:
                    return True
                version = importlib_metadata.version(dist_name)
                _required_dists.add('%s==%s' % (dist_name, version))
            except importlib_metadata.PackageNotFoundError as e:
                return True
        return res
    else:
        return True

_old_module_reduce = cloudpickle.cloudpickle_fast._module_reduce
def _module_reduce_include_env(obj):
    if obj == os:
        if cloudpickle.cloudpickle_fast._should_pickle_by_reference(obj):
            return cloudpickle.cloudpickle_fast.dynamic_subimport, (obj.__name__, {
                'environ': os.environ,
                'environb': os.environb,
            })
    return _old_module_reduce(obj)

def deserialize(obj):
    return cloudpickle.loads(obj['pickled'])

# TODO: add strict mode that uses `pipdeptree` to exactly match
# dependencies of packages as well.
def serialize(obj, custom_modules=[], match_library_patch_version=False, mimic_environ=True, force_versions={}, model_name=None):
    pbv = cloudpickle.list_registry_pickle_by_value().copy()
    try:
        for m in custom_modules:
            if m not in pbv:
                cloudpickle.register_pickle_by_value(m)
        PKG_TO_DIST = None
        _required_dists.clear()
        cloudpickle.cloudpickle._is_registered_pickle_by_value = _is_registered_pickle_by_value
        if mimic_environ:
            cloudpickle.cloudpickle_fast.CloudPickler._dispatch_table[
                cloudpickle.cloudpickle_fast.types.ModuleType
            ] = _module_reduce_include_env
        pickled = cloudpickle.dumps(obj)
    finally:
        cloudpickle.cloudpickle._is_registered_pickle_by_value = _old_is_registered_pickle_by_value
        cloudpickle.cloudpickle_fast.CloudPickler._dispatch_table[
            cloudpickle.cloudpickle_fast.types.ModuleType
        ] = _old_module_reduce
        for m in custom_modules:
            if m not in pbv:
                cloudpickle.unregister_pickle_by_value(m)

    reqs = list(_required_dists)
    reqs.sort()

    possible_names = set([])
    caller_frame = inspect.stack()[2].frame

    if model_name is None:
        # Derive name for model
        varmaps = [caller_frame.f_locals, caller_frame.f_globals]
        for varmap in varmaps:
            for k, v in varmap.items():
                if v is obj:
                    possible_names.add(k)

        if len(possible_names) == 0:
            raise RuntimeError('Unable to infer `model_name`, please specify explicitly.')
        elif len(possible_names) > 1:
            raise RuntimeError(f'Too many possible values for `model_name` '
                            + f'({possible_names}), please specify explicitly.')
        else:
            model_name = possible_names.pop()

    v = sys.version_info
    pyver = f'{v.major}.{v.minor}'
    if match_library_patch_version:
        pyver += f'.{v.micro}'
    else:
        for i in range(len(reqs)):
            reqs[i] = '.'.join(reqs[i].split('.')[:-1] + ['*'])

    reqs = [req for req in reqs if req.split('==')[0] not in force_versions]
    reqs += [f'{k}=={v}' for k, v in force_versions.items()]
    res = {
        'model_name': model_name,
        'pyver': pyver,
        'reqs': reqs,
    }
    return res, pickled



import pprint

class ServerError(RuntimeError):
    def __init__(self, code, msg, resp):
        super().__init__(f'({code}) {msg}')
        self.code = code
        self.msg = msg
        self.resp = resp

class RemoteCallError(RuntimeError):
    def __init__(self, server, resp):
        super().__init__(resp['exc_str'])
        self.exc_str = resp['exc_str']
        self.exc_type = resp['exc_type']
        self.traceback = resp['traceback']
        self.formatted = resp['formatted']
        self.server = server

    def __str__(self):
        return f'from {self.server}\nRemote Traceback (most recent call last):\n' + ''.join(self.formatted)

def json_or_err(resp):
    try:
        obj = resp.json
        if callable(obj):
            obj = obj()
    except Exception as e:
        raise ServerError(500, f'Server did not return JSON: {resp.data}', resp)

    if resp.status_code != 200:
        raise ServerError(resp.status_code, obj.get('err', obj), resp)

    return obj






import copy
import requests
import json
import sys
import base64
import time

from . import config, thunk, errors
from . import json_utils

# We pull these into the toplevel, but the module resolution can't
# know that.
SHORTENS = {
    'client.Client': 'Client',
    'client.Models': 'Models',
}
def repr_classname(obj):
    classname = obj.__class__.__qualname__
    maybe_module = getattr(obj.__class__, '__module__')
    if maybe_module is not None:
        classname = maybe_module + '.' + classname
    for k, v in SHORTENS.items():
        if classname.endswith(k):
            classname = classname[:-len(k)] + v
            break
    return classname

class Censored():
    def __repr__(self):
        return '...'
def repr_with_args(obj, kwargs):
    # We don't want API keys accidentally ending up in logs.
    if 'api_key' in kwargs and kwargs['api_key'][0] != kwargs['api_key'][1]:
        kwargs['api_key'] = (Censored(), None)
    argstr = ', '.join(f'{k}={repr(v[0])}' for k, v in kwargs.items() if v[0] != v[1])
    return f'{repr_classname(obj)}({argstr})'

# We have this extra level of abstraction to make life easier on the testing harness.
class Server():
    pass

class HttpServer(Server):
    def __init__(self, prefix_path):
        self.prefix_path = prefix_path

    def post(self, endpoint, body_json, auth):
        return requests.post(
            self.prefix_path+endpoint,
            headers={
                "Content-Type": 'application/json',
            },
            data=body_json,
            auth=auth,
        )

    def upload_blob(self, endpoint, blob, auth):
        return requests.post(
            self.prefix_path+endpoint,
            data=blob,
            auth=auth,
        )

    def get(self, endpoint, auth):
        return requests.get(
            self.prefix_path+endpoint,
            headers={"Content-Type": 'application/json'},
            auth=auth,
        )

    def __str__(self):
        return self.prefix_path

    def __repr__(self):
        return f'{repr_classname(self)}({repr(self.prefix_path)})'

class EndpointWrapper():
    def __init__(self, server):
        if isinstance(server, str):
            self.server = HttpServer(server)
        else:
            self.server = server

    def post(self, endpoint, obj):
        res = self.server.post(endpoint, obj, self.auth())
        return errors.json_or_err(res)

    def upload_blob(self, endpoint, blob):
        res = self.server.upload_blob(endpoint, blob, self.auth())
        return errors.json_or_err(res)

    def get(self, endpoint):
        res = self.server.get(endpoint, self.auth())
        return errors.json_or_err(res)

    def auth(self):
        return (self.api_key, '')

class CachedRepr:
    def __init__(self, _repr):
        self.__dmn_repr__ = _repr

    def __repr__(self):
        return self.__dmn_repr__

class Models(EndpointWrapper):
    def __init__(self, api_key=None, model_server=None, serialization_fallback=None,
                 stdout=sys.stdout, stderr=sys.stderr):
        super().__init__(config.get_default('model_server', model_server, 'Models'))
        self.api_key = config.get_default('api_key', api_key, 'Models')
        self.namespace = 'default'
        self.serialization_fallback = serialization_fallback
        self.stdout = stdout
        self.stderr = stderr
        self._repr = repr_with_args(self, {
            'api_key': (api_key, None),
            'model_server': (model_server, None),
            'serialization_fallback': (serialization_fallback, None),
            'stdout': (stdout, sys.stdout),
            'stderr': (stderr, sys.stderr),
        })

    def __call__(self, namespace='default'):
        if namespace == 'default':
            return NamespacedModels(
                f'{self.__repr__()}()',
                models=self,
                namespace=namespace,
            )
        else:
            return NamespacedModels(
                f'{self.__repr__()}({repr(namespace)})',
                models=self,
                namespace=namespace,
            )


    def _call_remote(self, namespace, model_name, obj, **kw):
        endpoint = f'v1/call/{namespace}/{model_name}'
        if 'version' in kw:
            endpoint += f':{kw["version"]}'

        body_json = json.dumps(obj)

        res = self.post(endpoint, body_json)
        if res['success']:
            if 'stdout' in res:
                bstr = base64.b64decode(res['stdout']['bytes'])
                if hasattr(self.stdout, 'buffer'):
                    self.stdout.buffer.write(bstr)
                    self.stdout.buffer.flush()
                else:
                    self.stdout.write(bstr.decode('utf-8', errors='replace'))
                    self.stdout.flush()
            if 'stderr' in res:
                bstr = base64.b64decode(res['stderr']['bytes'])
                if hasattr(self.stderr, 'buffer'):
                    self.stderr.buffer.write(bstr)
                    self.stderr.buffer.flush()
                else:
                    self.stderr.write(bstr.decode('utf-8', errors='replace'))
                    self.stderr.flush()
            return json_utils.loads(res['res'])
        else:
            raise errors.RemoteCallError(f'{self.server}{endpoint}', res)

    def __repr__(self):
        return self._repr

class NamespacedModels(CachedRepr):
    def __init__(self, _repr, **kw):
        super().__init__(_repr)
        self.__dmn__ = kw

    def __getattr__(self, model_name):
        return ModelView(
            f'{self.__repr__()}.{model_name}',
            model_name=model_name,
            method=[],
            **self.__dmn__,
        )

class VersionedModelView(CachedRepr):
    def __init__(self, _repr, **kw):
        super().__init__(_repr)
        self.__dmn__ = kw
        self.__dmn__.setdefault('methodless_repr', _repr)

    def __call__(self, *a, **kw):
        sf = self.__dmn__['models'].serialization_fallback
        return self.__dmn__['models']._call_remote(
            obj={
                'args': json_utils.dumps(a, serialization_fallback=sf),
                'kwargs': json_utils.dumps(kw, serialization_fallback=sf),
                'method': self.__dmn__['method'],
            },
            **self.__dmn__,
        )

    def __getattr__(self, method_part):
        no_method = self.__dmn__.copy()
        del no_method['method']
        return VersionedModelView(
            f'{self.__repr__()}.{method_part}',
            method=(self.__dmn__['method']+[method_part]),
            **no_method,
        )

class ModelView(VersionedModelView):
    def __init__(self, _repr, **kw):
        super().__init__(_repr, **kw)

    def __getitem__(self, version):
        return VersionedModelView(
            f'{self.__repr__()}[{repr(version)}]',
            version=version,
            **self.__dmn__,
        )

class Client(EndpointWrapper):
    def __init__(self, api_key=None, admin_server=None, model_server=None, app_server=None, serialization_fallback=None, pre_upload=True, build_mode='kaniko', stdout=sys.stdout, stderr=sys.stderr):
        super().__init__(config.get_default('admin_server', admin_server, 'Client'))
        self.api_key = config.get_default('api_key', api_key, 'Client')
        self.models = Models(self.api_key, model_server=model_server, serialization_fallback=serialization_fallback, stdout=stdout, stderr=stderr)
        self.app_server = config.get_default('app_server', app_server, 'Client')
        self.pre_upload = pre_upload
        self.build_mode = build_mode
        self._repr = repr_with_args(self, {
            'api_key': (api_key, None),
            'admin_server': (admin_server, None),
            'model_server': (model_server, None),
            'serialization_fallback': (serialization_fallback, None),
            'pre_upload': (pre_upload, True),
            'build_mode': (build_mode, 'kaniko'),
            'stdout': (stdout, sys.stdout),
            'stderr': (stderr, sys.stderr),
        })
        self.models._repr = self._repr + '.models'

    def deploy_status(self, master, version):
        return self.post('v1/deploy_status',
                         json.dumps({'master': master, 'version': version}))

    # TODO RSI: allow the same __call__ interface for adding a namespace.
    def deploy(self, obj, namespace='default', model_name=None, serialize_opts={}, quiet=False):
        if isinstance(obj, VersionedModelView):
            if obj.__dmn__['method'] != []:
                raise RuntimeError(
                    'Only top-level models can be redeployed.  ' +
                    f'Instead of deploying {repr(obj)}, deploy ' +
                    f'{obj.__dmn__["methodless_repr"]}.')
            res = self.post(
                f'v1/redeploy',
                json.dumps({
                    'namespace': namespace,
                    'model_name': model_name or obj.__dmn__['model_name'],

                    'from_namespace': obj.__dmn__['namespace'],
                    'from_model_name': obj.__dmn__['model_name'],
                    'from_version': obj.__dmn__.get('version', None),
                }),
            )
        else:
            deploy_args, pickled = thunk.serialize(
                obj, model_name=model_name, **serialize_opts)
            deploy_args['namespace'] = namespace
            if self.build_mode is not None:
                deploy_args['build_mode'] = self.build_mode
            if self.pre_upload:
                res = self.upload_blob('v1/upload', pickled)
                deploy_args['pickled_ref'] = res['ref']
            else:
                deploy_args['pickled_b64'] = base64.encodebytes(pickled).decode('utf-8')
            res = self.post('v1/deploy', json.dumps(deploy_args))

        deploy_status = DeployStatus(self, res)
        if not quiet and deploy_status.url is not None:
            print("Deploy in progress.  Watch here: " + deploy_status.url)
        return deploy_status

    def __repr__(self):
        return self._repr

    def endpoints(self):
        return self.get('v1/endpoints')

class DeployStatus:
    def __init__(self, client, data, app_server=None):
        self.client = client
        self.data = data
        self.as_of = time.time()

        self.url = None
        if self.data.get('deploy', None) is not None:
            self.url = f"{self.client.app_server}deploy_status/{self.data['deploy']['master']}/{self.data['deploy']['version']}"

    def __repr__(self):
        return f'DeployStatus(as_of={self.as_of}, data={self.data}, url={repr(self.url)})'
    def has_condition(self, condition):
        return condition in self.data['conditions']

    def refresh(self):
        self.as_of = time.time()
        self.data = self.client.deploy_status(**self.data['deploy'])
        return self

    def wait(self, condition='ready', error_if_superseded=True,
             timeout=3600, sleep_time=1, max_sleep_time=300):
        if condition is None:
            condition = 'ready'
        legal_conditions = ['ready', 'stable']
        if condition not in legal_conditions:
            raise RuntimeError(
                f'`condition` must be in `{legal_conditions}`, got `{condition}`.')

        start_time = time.time()
        cur_sleep_time = sleep_time
        while not self.has_condition(condition):
            self.refresh()
            if self.has_condition('failed'):
                raise RuntimeError(f'Deploy failed: {self}')
            if self.has_condition('superseded'):
                if error_if_superseded:
                    raise RuntimeError(
                        f'Deploy superseded by a later deploy: {self}')
            if time.time() > start_time + timeout:
                raise TimeoutError(f'Timed out waiting for condition `{condition}`.')
            time.sleep(cur_sleep_time)
            cur_sleep_time = min(1.5*cur_sleep_time, max_sleep_time)

        return self



import os
import sys
import json

_config_vars = {
    'DEBUG': False,
    'CONFIG_PATH': os.path.expanduser('~/.config/carbondeploy/client.json'),
    'ADMIN_SERVER': 'https://admin.carbondeploy.com/',
    'MODEL_SERVER': 'https://models.carbondeploy.com/',
    'APP_SERVER_BACKEND': 'https://app.carbondeploy.com/',
    'APP_SERVER': 'https://www.carbondeploy.com/',
    'ORG': None,
    'USERNAME': '',
    'PASSWORD': None,
    'API_KEY': None,
}

for v, val in _config_vars.items():
    globals()[v] = val

def env_name(var):
    return 'DMN_'+var.upper()
_source = {}

for v in _config_vars.keys():
    if env_name(v) in os.environ:
        globals()[v] = os.environ[env_name(v)]
        _source[v] = 'env'

if os.path.exists(CONFIG_PATH):
    try:
        with open(CONFIG_PATH, 'r') as f:
            CONFIG = json.load(f)
        for v, val in CONFIG.items():
            v = v.upper()
            if v not in _config_vars:
                raise RuntimeError('Unrecognized config option `%s`.  Options are: %s.' %
                                   (v, list(CONFIG.keys())))
            if v not in _source:
                globals()[v] = val
                _source[v] = 'config_file'
    except Exception as e:
        raise RuntimeError('Unable to load config file `%s`: %s' % (CONFIG_PATH, e))

def get_default(_v, default_val, classname):
    if default_val is not None:
        return default_val
    v = _v.upper()
    val = globals().get(v)
    if val is None:
        raise RuntimeError(f'No {_v} specified.  Either pass it as an argument to `{classname}`, ' +
                           f'in the environment variable `{env_name(_v)}`, ' +
                           f'or add `{_v}` to the config file at `{CONFIG_PATH}`.')
    return val



from setuptools import setup

setup(name='carbondeploy',
      version='1.0.0',
      description='Library for deploying Jupyter notebooks.',
      url='https://www.carbondeploy.com/',
      author='Michael Lucy',
      author_email='michaelglucy@gmail.com',
      license='MIT',
      packages=['carbondeploy'],
      zip_safe=True,

      install_requires=[
          'requests>=2.24.0',
          'cloudpickle==2.0.0',
          'importlib-metadata>=4.8.3',
      ],
      python_requires='>=3',
)
